package eu.mhsl.craftattack.spawn.common.appliances.metaGameplay.settings;

import eu.mhsl.craftattack.spawn.core.Main;
import eu.mhsl.craftattack.spawn.core.appliance.Appliance;
import eu.mhsl.craftattack.spawn.core.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.common.appliances.metaGameplay.settings.datatypes.Setting;
import eu.mhsl.craftattack.spawn.common.appliances.metaGameplay.settings.listeners.OpenSettingsShortcutListener;
import eu.mhsl.craftattack.spawn.common.appliances.metaGameplay.settings.listeners.SettingsInventoryListener;
import net.kyori.adventure.text.Component;
import org.bukkit.Bukkit;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.bukkit.inventory.Inventory;
import org.jetbrains.annotations.NotNull;

import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

public class Settings extends Appliance {
    private static Settings settingsInstance;
    private final Set<Class<? extends Setting<?>>> declaredSettings = new HashSet<>();

    public enum Key {
        TechnicalTab,
        ShowJoinAndLeaveMessages,
        EnablePortableCrafting,
        EnableSettingsShortcut,
        AutoShulker,
        SignEdit,
        HotbarReplacer,
        ChatMentions,
        DoubleDoors,
        KnockDoors,
        BorderWarning
    }

    public static Settings instance() {
        if(settingsInstance != null) return settingsInstance;
        Settings.settingsInstance = Main.instance().getAppliance(Settings.class);
        return settingsInstance;
    }

    public void declareSetting(Class<? extends Setting<?>> setting) {
        this.declaredSettings.add(setting);
        this.settingsCache.clear();
    }

    @Override
    public void onEnable() {
        Settings.instance().declareSetting(SettingsShortcutSetting.class);
    }

    public record OpenSettingsInventory(Inventory inventory, List<Setting<?>> settings) {
    }

    private final WeakHashMap<Player, OpenSettingsInventory> openSettingsInventories = new WeakHashMap<>();
    private final WeakHashMap<Player, List<Setting<?>>> settingsCache = new WeakHashMap<>();

    private List<Setting<?>> getSettings(Player player) {
        if(this.settingsCache.containsKey(player)) return this.settingsCache.get(player);

        List<Setting<?>> settings = this.declaredSettings.stream()
            .map(clazz -> {
                try {
                    return clazz.getDeclaredConstructor();
                } catch(NoSuchMethodException e) {
                    throw new RuntimeException(String.format("Setting '%s' does not have an accessible constructor", clazz.getName()), e);
                }
            })
            .map(constructor -> {
                try {
                    return constructor.newInstance();
                } catch(InstantiationException | IllegalAccessException | InvocationTargetException e) {
                    throw new RuntimeException(String.format("Failed to create instance of setting '%s'", constructor.getDeclaringClass().getName()), e);
                }
            })
            .peek(setting -> setting.initializeFromPlayer(player))
            .collect(Collectors.toList());

        this.settingsCache.put(player, settings);
        return settings;
    }

    public <T> T getSetting(Player player, Key key, Class<T> clazz) {
        Setting<?> setting = this.getSettings(player).stream()
            .filter(s -> Objects.equals(s.getKey(), key))
            .findFirst()
            .orElseThrow();

        if(!clazz.equals(setting.dataType()))
            throw new IllegalStateException("Tried to retrieve Setting with Datatype " + clazz.getSimpleName() + " but expected " + setting.dataType().getSimpleName());
        if(!clazz.isInstance(setting.state()))
            throw new ClassCastException(clazz.getSimpleName() + " is not an instance of " + setting.dataType().getSimpleName());
        return clazz.cast(setting.state());
    }

    public void openSettings(Player player) {
        List<Setting<?>> settings = this.getSettings(player);
        Inventory inventory = Bukkit.createInventory(null, this.calculateInvSize(settings), Component.text("Einstellungen"));

        AtomicInteger row = new AtomicInteger(0);
        Arrays.stream(SettingCategory.values())
            .forEach(category -> {
                List<Setting<?>> categorizedSettings = settings.stream()
                    .filter(setting -> setting instanceof CategorizedSetting)
                    .filter(setting -> ((CategorizedSetting) setting).category().equals(category))
                    .toList();

                //skip empty category rows
                if(categorizedSettings.isEmpty()) return;

                for(int i = 0; i < categorizedSettings.size(); i++) {
                    int slot = row.get() * 9 + i % 9;
                    inventory.setItem(slot, categorizedSettings.get(i).buildItem());

                    if(i % 9 == 8) {
                        row.incrementAndGet();
                    }
                }
                row.incrementAndGet();
            });

        List<Setting<?>> uncategorizedSettings = settings.stream()
            .filter(setting -> !(setting instanceof CategorizedSetting))
            .toList();

        for(int i = 0; i < uncategorizedSettings.size(); i++) {
            int slot = row.get() * 9 + i % 9;
            inventory.setItem(slot, uncategorizedSettings.get(i).buildItem());

            if(i % 9 == 8) {
                row.incrementAndGet();
            }
        }

        player.openInventory(inventory);
        this.openSettingsInventories.put(player, new OpenSettingsInventory(inventory, settings));
    }

    private int calculateInvSize(List<Setting<?>> settings) {
        int countOfUncategorized = (int) settings.stream()
            .filter(setting -> !(setting instanceof CategorizedSetting))
            .count();
        int invSizeForUncategorized = (int) Math.ceil((double) countOfUncategorized / 9) * 9;

        int invSizeForCategorized = Arrays.stream(SettingCategory.values())
            .map(settingCategory -> settings.stream()
                .filter(setting -> setting instanceof CategorizedSetting)
                .map(setting -> (CategorizedSetting) setting)
                .filter(categorizedSetting -> categorizedSetting.category().equals(settingCategory))
                .count())
            .map(itemCount -> (int) Math.ceil((double) itemCount / 9))
            .reduce(Integer::sum)
            .orElse(1) * 9;

        int invSize = invSizeForUncategorized + invSizeForCategorized;
        if(invSize % 9 != 0) throw new IllegalStateException(
            String.format("Failed to calculate settings inventory size. %d is not an multiple of 9", invSize)
        );
        return invSize;
    }

    public void onSettingsClose(Player player) {
        if(!this.openSettingsInventories.containsKey(player)) return;
        this.openSettingsInventories.remove(player);
        player.updateInventory();
    }

    public boolean hasSettingsNotOpen(Player player) {
        return !this.openSettingsInventories.containsKey(player);
    }

    public OpenSettingsInventory getOpenInventory(Player player) {
        if(this.hasSettingsNotOpen(player))
            throw new RuntimeException("Cannot retrieve data from closed Settings inventory!");
        return this.openSettingsInventories.get(player);
    }

    @Override
    @NotNull
    protected List<Listener> listeners() {
        return List.of(
            new SettingsInventoryListener(),
            new OpenSettingsShortcutListener()
        );
    }

    @Override
    @NotNull
    protected List<ApplianceCommand<?>> commands() {
        return List.of(new SettingsCommand());
    }
}
