package eu.mhsl.craftattack.spawn.core;

import eu.mhsl.craftattack.spawn.core.api.client.RepositoryLoader;
import eu.mhsl.craftattack.spawn.core.api.server.HttpServer;
import eu.mhsl.craftattack.spawn.core.appliance.Appliance;
import eu.mhsl.craftattack.spawn.core.config.Configuration;
import org.bukkit.Bukkit;
import org.bukkit.event.HandlerList;
import org.bukkit.plugin.java.JavaPlugin;
import org.reflections.Reflections;

import java.lang.reflect.ParameterizedType;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class Main extends JavaPlugin {
    public static final String projectPackage = "eu.mhsl.craftattack.spawn";
    private static Main instance;
    private static Logger logger;

    private List<Appliance> appliances;
    private RepositoryLoader repositoryLoader;
    private Reflections reflections;

    @Override
    public void onEnable() {
        instance = this;
        logger = instance().getLogger();
        this.saveDefaultConfig();
        try {
            this.wrappedEnable();
        } catch(Exception e) {
            Main.logger().log(Level.SEVERE, "Error while initializing Spawn plugin, shutting down!", e);
            Bukkit.shutdown();
        }
    }

    private void wrappedEnable() {
        Configuration.readConfig();
        List<String> disabledAppliances = Configuration.pluginConfig.getStringList("disabledAppliances");

        Main.logger().info("Initializing reflections...");
        this.reflections = new Reflections(projectPackage);

        Main.logger().info("Loading Repositories...");
        this.repositoryLoader = new RepositoryLoader();
        Main.logger().info(String.format("Loaded %d repositories!", this.repositoryLoader.getRepositories().size()));

        Main.logger().info("Loading appliances...");
        this.appliances = this.findSubtypesOf(Appliance.class).stream()
            .filter(applianceClass -> !disabledAppliances.contains(applianceClass.getSimpleName()))
            .map(applianceClass -> {
                try {
                    return (Appliance) applianceClass.getDeclaredConstructor().newInstance();
                } catch(Exception e) {
                    throw new RuntimeException(String.format("Failed to create instance of '%s'", applianceClass.getName()), e);
                }
            })
            .toList();
        Main.logger().info(String.format("Loaded %d appliances!", this.appliances.size()));

        Main.logger().info("Initializing appliances...");
        this.appliances.forEach(appliance -> {
            appliance.onEnable();
            appliance.initialize(this);
        });
        Main.logger().info(String.format("Initialized %d appliances!", this.appliances.size()));

        if(Configuration.pluginConfig.getBoolean("httpServerEnabled", true)) {
            Main.logger().info("Starting HTTP API...");
            new HttpServer();
        }

        this.getServer().getMessenger().registerOutgoingPluginChannel(this, "BungeeCord");
        Main.logger().info("Startup complete!");
    }

    @Override
    public void onDisable() {
        Main.logger().info("Disabling appliances...");
        this.appliances.forEach(appliance -> {
            Main.logger().info("Disabling " + appliance.getClass().getSimpleName());
            appliance.onDisable();
            appliance.destruct(this);
        });

        HandlerList.unregisterAll(this);
        Bukkit.getScheduler().cancelTasks(this);
        Main.logger().info("Disabled " + this.appliances.size() + " appliances!");
    }

    public <T extends Appliance> T getAppliance(Class<T> clazz) {
        return this.appliances.stream()
            .filter(clazz::isInstance)
            .map(clazz::cast)
            .findFirst()
            .orElseThrow(() -> new RuntimeException(String.format("Appliance %s not loaded or instantiated!", clazz)));
    }

    public <T> Set<Class<? extends T>> findSubtypesOf(Class<T> type) {
        return this.reflections.getSubTypesOf(type);
    }

    @SuppressWarnings("unchecked")
    public static <T> Class<T> getApplianceType(Class<?> clazz) {
        return (Class<T>) ((ParameterizedType) clazz.getGenericSuperclass()).getActualTypeArguments()[0];
    }

    public List<Appliance> getAppliances() {
        return this.appliances;
    }

    public RepositoryLoader getRepositoryLoader() {
        return this.repositoryLoader;
    }

    public static Main instance() {
        return instance;
    }

    public static Logger logger() {
        return logger;
    }
}
