package eu.mhsl.craftattack.spawn.core.api.client;

import eu.mhsl.craftattack.spawn.core.Main;
import org.apache.http.client.utils.URIBuilder;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

@RepositoryLoader.Abstraction
public abstract class HttpRepository extends Repository {
    public record RequestModifier(
        @Nullable Consumer<URIBuilder> uri,
        @Nullable Consumer<HttpRequest.Builder> header
    ) {}

    private final List<RequestModifier> baseRequestModifier;

    public HttpRepository(URI basePath) {
        this(basePath, new RequestModifier(null, null));
    }

    public HttpRepository(URI basePath, RequestModifier... baseRequestModifier) {
        super(basePath);

        this.baseRequestModifier = baseRequestModifier == null
            ? List.of()
            : List.of(baseRequestModifier);
    }

    protected <TInput, TOutput> ReqResp<TOutput> post(String command, TInput data, Class<TOutput> outputType) {
        return this.post(command, parameters -> {
        }, data, outputType);
    }

    protected <TInput, TOutput> ReqResp<TOutput> post(String command, Consumer<URIBuilder> parameters, TInput data, Class<TOutput> outputType) {
        HttpRequest request = this.getRequestBuilder(this.getUri(command, parameters))
            .POST(HttpRequest.BodyPublishers.ofString(this.gson.toJson(data)))
            .build();

        return this.execute(request, outputType);
    }

    protected <TInput, TOutput> ReqResp<TOutput> put(String command, TInput data, Class<TOutput> outputType) {
        return this.put(command, parameters -> {
        }, data, outputType);
    }

    protected <TInput, TOutput> ReqResp<TOutput> put(String command, Consumer<URIBuilder> parameters, TInput data, Class<TOutput> outputType) {
        HttpRequest request = this.getRequestBuilder(this.getUri(command, parameters))
            .PUT(HttpRequest.BodyPublishers.ofString(this.gson.toJson(data)))
            .build();

        return this.execute(request, outputType);
    }

    protected <TOutput> ReqResp<TOutput> get(String command, Class<TOutput> outputType) {
        return this.get(command, parameters -> {
        }, outputType);
    }

    protected <TOutput> ReqResp<TOutput> get(String command, Consumer<URIBuilder> parameters, Class<TOutput> outputType) {
        HttpRequest request = this.getRequestBuilder(this.getUri(command, parameters))
            .GET()
            .build();

        return this.execute(request, outputType);
    }

    private URI getUri(String command, Consumer<URIBuilder> parameters) {
        try {
            URIBuilder builder = new URIBuilder(this.basePath + "/" + command);
            this.baseRequestModifier.stream()
                .map(requestModifier -> requestModifier.uri)
                .filter(Objects::nonNull)
                .forEach(modifier -> modifier.accept(builder));

            parameters.accept(builder);
            return builder.build();
        } catch(URISyntaxException e) {
            throw new RuntimeException(e);
        }
    }

    private HttpRequest.Builder getRequestBuilder(URI endpoint) {
        HttpRequest.Builder builder = HttpRequest.newBuilder()
            .uri(endpoint)
            .header("User-Agent", Main.instance().getServer().getBukkitVersion())
            .header("Content-Type", "application/json");

        this.baseRequestModifier.stream()
            .map(requestModifier -> requestModifier.header)
            .filter(Objects::nonNull)
            .forEach(modifier -> modifier.accept(builder));

        return builder;
    }

    private <TResponse> ReqResp<TResponse> execute(HttpRequest request, Class<TResponse> clazz) {
        ReqResp<String> rawResponse = this.sendHttp(request);
        Main.logger().info(String.format(
            "Request: %s\nResponse: %s",
            request,
            rawResponse
        ));
        return new ReqResp<>(rawResponse.status(), this.gson.fromJson(rawResponse.data(), clazz));
    }

    private ReqResp<String> sendHttp(HttpRequest request) {
        try(HttpClient client = HttpClient.newHttpClient()) {
            this.validateThread(request.uri().getPath());
            HttpResponse<String> httpResponse = client.send(request, HttpResponse.BodyHandlers.ofString());
            return new ReqResp<>(httpResponse.statusCode(), httpResponse.body());
        } catch(IOException | InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

}
