package eu.mhsl.craftattack.spawn.core.appliance;

import eu.mhsl.craftattack.spawn.core.Main;
import eu.mhsl.craftattack.spawn.core.api.client.Repository;
import eu.mhsl.craftattack.spawn.core.api.server.HttpServer;
import eu.mhsl.craftattack.spawn.core.config.Configuration;
import org.bukkit.Bukkit;
import org.bukkit.command.PluginCommand;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.event.HandlerList;
import org.bukkit.event.Listener;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.NotNull;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * Any implementation of this class can be seen as a "sub-plugin" with its own event handlers and commands.
 * Appliances can be enabled or disabled independent of other appliances
 */
public abstract class Appliance {
    private String localConfigPath;
    private List<Listener> listeners;
    private List<ApplianceCommand<?>> commands;

    public Appliance() {
    }

    /**
     * Use this constructor to specify a config sub-path for use with the localConfig() method.
     *
     * @param localConfigPath sub path, if not found, the whole config will be used
     */
    public Appliance(String localConfigPath) {
        this.localConfigPath = localConfigPath;
    }

    /**
     * Provides a list of listeners for the appliance. All listeners will be automatically registered.
     *
     * @return List of listeners
     */
    @NotNull
    protected List<Listener> listeners() {
        return new ArrayList<>();
    }

    /**
     * Provides a list of commands for the appliance. All commands will be automatically registered.
     *
     * @return List of commands
     */
    @NotNull
    protected List<ApplianceCommand<?>> commands() {
        return new ArrayList<>();
    }

    /**
     * Called on initialization to add all needed API Routes.
     * The routeBuilder can be used to get the correct Path prefixes
     *
     * @param apiBuilder holds data for needed route prefixes.
     */
    public void httpApi(HttpServer.ApiBuilder apiBuilder) {
    }

    /**
     * Provides a localized config section. Path can be set in appliance constructor.
     *
     * @return Section of configuration for your appliance
     */
    @NotNull
    public ConfigurationSection localConfig() {
        return Optional.ofNullable(Configuration.cfg.getConfigurationSection(this.localConfigPath))
            .orElseGet(() -> Configuration.cfg.createSection(this.localConfigPath));
    }

    public void onEnable() {
    }

    public void onDisable() {
    }

    public void initialize(@NotNull JavaPlugin plugin) {
        this.listeners = this.listeners();
        this.commands = this.commands();

        this.listeners.forEach(listener -> Bukkit.getPluginManager().registerEvents(listener, plugin));
        this.commands.forEach(command -> this.setCommandExecutor(plugin, command.commandName, command));
    }

    public void destruct(@NotNull JavaPlugin plugin) {
        this.listeners.forEach(HandlerList::unregisterAll);
    }

    protected <T extends Appliance> T queryAppliance(Class<T> clazz) {
        return Main.instance().getAppliance(clazz);
    }

    protected <T extends Repository> T queryRepository(Class<T> clazz) {
        return Main.instance().getRepositoryLoader().getRepository(clazz);
    }

    private void setCommandExecutor(JavaPlugin plugin, String name, ApplianceCommand<?> executor) {
        try {
            PluginCommand command = this.createPluginCommand(name, plugin);
            command.setExecutor(executor);
            command.setTabCompleter(executor);
            plugin.getServer().getCommandMap().register(plugin.getName(), command);
        } catch(Exception e) {
            throw new RuntimeException(String.format("Failed to register command '%s'", name), e);
        }
    }

    private PluginCommand createPluginCommand(String name, JavaPlugin plugin) throws Exception {
        Constructor<PluginCommand> constructor = PluginCommand.class.getDeclaredConstructor(String.class, Plugin.class);
        constructor.setAccessible(true);
        return constructor.newInstance(name, plugin);
    }

    public List<Listener> getListeners() {
        return this.listeners;
    }

    public List<ApplianceCommand<?>> getCommands() {
        return this.commands;
    }
}
