package eu.mhsl.craftattack.spawn.varo.appliances.metaGameplay.teams;

import eu.mhsl.craftattack.spawn.core.Main;
import eu.mhsl.craftattack.spawn.core.appliance.Appliance;
import eu.mhsl.craftattack.spawn.core.appliance.ApplianceCommand;
import eu.mhsl.craftattack.spawn.core.util.text.Countdown;
import eu.mhsl.craftattack.spawn.core.util.text.DisconnectInfo;
import eu.mhsl.craftattack.spawn.common.appliances.metaGameplay.displayName.DisplayName;
import eu.mhsl.craftattack.spawn.varo.api.repositories.TeamRepository;
import eu.mhsl.craftattack.spawn.varo.appliances.internal.teamTasks.TeamTasks;
import eu.mhsl.craftattack.spawn.varo.appliances.internal.teamTasks.tasks.CountdownTeamTask;
import eu.mhsl.craftattack.spawn.varo.appliances.metaGameplay.joinProtection.JoinProtection;
import eu.mhsl.craftattack.spawn.varo.appliances.metaGameplay.playTimer.PlayTimer;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import net.kyori.adventure.text.format.TextColor;
import net.kyori.adventure.util.Ticks;
import org.bukkit.Bukkit;
import org.bukkit.OfflinePlayer;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

public class Teams extends Appliance implements DisplayName.Prefixed {
    private final List<VaroTeam> teams = new ArrayList<>();

    public Teams() {
        Bukkit.getScheduler().runTaskTimerAsynchronously(
            Main.instance(),
            this::refreshTeamList,
            0,
            Ticks.TICKS_PER_SECOND * 60
        );
    }

    public void refreshTeamList() {
        var updatedTeams = this.queryRepository(TeamRepository.class).getTeams().data();

        for (var updatedTeam : updatedTeams) {
            VaroTeam existingTeam = this.findTeamByName(updatedTeam.name());

            if (existingTeam != null) {
                existingTeam.members = updatedTeam.users().stream()
                    .map(user -> new VaroTeam.Member(user.uuid(), user.dead()))
                    .toList();
                existingTeam.color = updatedTeam.color();
                existingTeam.name = updatedTeam.name();
            } else {
                VaroTeam newTeam = new VaroTeam(
                    updatedTeam.users().stream()
                        .map(user -> new VaroTeam.Member(user.uuid(), user.dead()))
                        .toList(),
                    updatedTeam.name(),
                    updatedTeam.color()
                );
                this.teams.add(newTeam);
                Main.logger().info("Added missing team to Teams registry: " + newTeam.name);
            }
        }
    }

    public @Nullable VaroTeam findTeamByName(String name) {
        for (VaroTeam team : this.teams) {
            if (team.name.equals(name)) return team;
        }
        return null;
    }

    public boolean canLogin(UUID playerId) {
        OfflinePlayer offlinePlayer = Bukkit.getOfflinePlayer(playerId);
        if(offlinePlayer.isOp()) {
            Main.logger().info(String.format("Allowing player %s to login, because he ist OP!", playerId));
            return true;
        }

        Optional<VaroTeam> team = this.teams.stream()
            .filter(varoTeam -> varoTeam.hasMember(playerId) && !Objects.requireNonNull(varoTeam.getMemberById(playerId)).isDead)
            .findAny();

        team.ifPresentOrElse(
            found -> Main.logger().info(String.format("Player %s is in Team %s!", playerId, found.name)),
            () -> Main.logger().info(String.format("No valid Team found for %s (or he is in a Team but dead)!", playerId))
        );
        return team.isPresent();
    }

    public @Nullable VaroTeam getTeamFromPlayer(UUID playerId) {
        return this.teams.stream()
            .filter(varoTeam -> varoTeam.hasMember(playerId))
            .findFirst()
            .orElse(null);
    }

    public List<VaroTeam> getAllTeams() {
        return this.teams;
    }

    public void enforceTeamJoin(Player joinedPlayer) {
        if(joinedPlayer.isOp()) return;

        DisconnectInfo teamNotCompleteInfo = new DisconnectInfo(
            "Teampartner nicht beigetreten",
            "Deine Verbindung wurde getrennt, da dein Teampartner keine Verbindung zum Server hergestellt hat!",
            "Bitte sorge dafür, dass alle anderen Teammitglieder eine einwandfreie Internetverbindung haben und melde dich im Zweifel bei einem Admin!",
            joinedPlayer.getUniqueId()
        );

        DisconnectInfo teamNoPlaytime = new DisconnectInfo(
            "Keine Spielzeit verfügbar",
            "Deine Verbindung wurde getrennt, da dein Team keine verbleibende Spielzeit auf dem Server hat!",
            "Falls dies ein Fehler ist, melde dich bitte bei einem Admin!",
            joinedPlayer.getUniqueId()
        );

        VaroTeam team = this.getTeamFromPlayer(joinedPlayer.getUniqueId());
        if(team == null) throw new IllegalStateException("Player must be in a Team");

        PlayTimer playTimer = Main.instance().getAppliance(PlayTimer.class);
        boolean isAllowed = playTimer.tryConsumeTicket(team);
        if(!isAllowed) {
            Main.logger().warning(String.format("Team %s joined, but got denied from Ticketing. Team will be kicked!", team.name));
            team.kickTeam(teamNoPlaytime);
            return;
        }

        int leftTickets = playTimer.getTickets(team);
        Main.logger().info(String.format("Player %s joined. There are %d tickets left!", joinedPlayer.getName(), leftTickets));
        String playtimeOverview = String.format(
            "Dein Team hat noch %d Beitritte, also %dx%d=%d Minuten übrig.",
            leftTickets,
            leftTickets,
            PlayTimer.PLAYTIME_MINUTES,
            leftTickets * PlayTimer.PLAYTIME_MINUTES
        );
        joinedPlayer.sendMessage(Component.text(
            leftTickets == 0
                ? String.format("Dein Team hat ab jetzt %d Minuten Spielzeit!", PlayTimer.PLAYTIME_MINUTES)
                : playtimeOverview,
            NamedTextColor.GREEN
        ));

        Bukkit.getScheduler().scheduleSyncDelayedTask(
            Main.instance(),
            () ->  team.members.stream()
                .filter(member -> !member.isDead)
                .filter(member -> {
                    Player p = Bukkit.getPlayer(member.player.getUniqueId());
                    return p == null || !p.isOnline();
                })
                .findAny()
                .ifPresentOrElse(
                    member -> team.kickTeam(teamNotCompleteInfo),
                    () -> {
                        Main.logger().info(String.format("Starting Time countdown for Team %s with %d", team.name, PlayTimer.PLAYTIME_MINUTES));
                        CountdownTeamTask countdown = new CountdownTeamTask(
                            60 * PlayTimer.PLAYTIME_MINUTES,
                            announcementData -> Component.text(
                                String.format("Es verbleiben noch %d %s Spielzeit!", announcementData.count(), announcementData.unit()),
                                NamedTextColor.RED
                            ),
                            component -> team.getOnlinePlayers().forEach(player -> player.sendMessage(component)),
                            team::timeOverKick
                        );
                        countdown.setDefaultAnnouncements(count -> {
                            if(count > 300) return null;
                            if(count > 60 && count % 60 == 0) {
                                return new Countdown.AnnouncementData(count / 60, "Minuten");
                            }

                            if(count <= 30 && (count <= 10 || count % 10 == 0)) {
                                return new Countdown.AnnouncementData(count, "Sekunden");
                            }

                            return null;
                        });
                        countdown.start();
                        this.queryAppliance(TeamTasks.class).addTask(team, TeamTasks.Type.TIME_KICK, countdown);
                    }
                ),
            Ticks.TICKS_PER_SECOND * (JoinProtection.resistanceDuration / 2)
        );
    }

    public void enforceTeamLeave(Player leftPlayer) {
        DisconnectInfo disconnectInfo = new DisconnectInfo(
            "Dein Teampartner hat den Server vorzeitig verlassen",
            "Dein Team erhält daher einen Strike!",
            "Bitte bleibt stets die vollen 30 Minuten auf dem Server!",
            leftPlayer.getUniqueId()
        );
        VaroTeam team = this.getTeamFromPlayer(leftPlayer.getUniqueId());
        if(team != null) team.kickTeam(disconnectInfo);
    }

    @Override
    public @Nullable Component getNamePrefix(Player player) {
        VaroTeam team = this.getTeamFromPlayer(player.getUniqueId());
        return Component.text(
            String.format("[%s]", team != null ? team.name : "?"),
            TextColor.fromCSSHexString(team != null ? team.color : "#ffffff")
        );
    }

    @Override
    protected @NotNull List<Listener> listeners() {
        return List.of(new ConnectivityChangeListener());
    }

    @Override
    protected @NotNull List<ApplianceCommand<?>> commands() {
        return List.of(new TeamListCommand());
    }
}
